;; Generated package description from dired-preview.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "dired-preview" "0.6.1.0.20260610.0" "Automatically preview file at point in Dired" '((emacs "28.1")) :commit "090eded80f801cb4f651490cdd9a78910ef63946" :authors '(("Protesilaos" . "info@protesilaos.com")) :maintainer '("Protesilaos" . "info@protesilaos.com") :keywords '("files" "convenience") :url "https://github.com/protesilaos/dired-preview")
